"""Command line interface for oak-ai."""
import codecs
import logging
import pickle
import sys
from copy import copy
from dataclasses import dataclass
from io import BytesIO, TextIOWrapper
from pathlib import Path
from typing import List, Optional

import click
import jsonlines
import openai
import yaml
from oaklib import get_adapter
from oaklib.cli import query_terms_iterator
from oaklib.io.streaming_csv_writer import StreamingCsvWriter

from ontogpt import __version__
from ontogpt.clients import OpenAIClient
from ontogpt.clients.pubmed_client import PubmedClient
from ontogpt.clients.soup_client import SoupClient
from ontogpt.clients.wikipedia_client import WikipediaClient
from ontogpt.engines import create_engine
from ontogpt.engines.embedding_similarity_engine import SimilarityEngine
from ontogpt.engines.enrichment import EnrichmentEngine
from ontogpt.engines.halo_engine import HALOEngine
from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.engines.synonym_engine import SynonymEngine
from ontogpt.evaluation.enrichment.eval_enrichment import EvalEnrichment
from ontogpt.evaluation.resolver import create_evaluator
from ontogpt.io.html_exporter import HTMLExporter
from ontogpt.io.markdown_exporter import MarkdownExporter
from ontogpt.utils.gene_set_utils import (
    GeneSet,
    _is_human,
    fill_missing_gene_set_values,
    parse_gene_set,
)

__all__ = [
    "main",
]

from ontogpt.io.owl_exporter import OWLExporter
from ontogpt.io.rdf_exporter import RDFExporter
from ontogpt.io.yaml_wrapper import dump_minimal_yaml
from ontogpt.templates.core import ExtractionResult


@dataclass
class Settings:
    """Global command line settings."""

    cache_db: Optional[str] = None
    skip_annotators: Optional[List[str]] = None


settings = Settings()


def _as_text_writer(f):
    if isinstance(f, TextIOWrapper):
        return f
    else:
        return codecs.getwriter("utf-8")(f)


def write_extraction(
    results: ExtractionResult,
    output: BytesIO,
    output_format: str = None,
    knowledge_engine: KnowledgeEngine = None,
):
    if output_format == "pickle":
        output.write(pickle.dumps(results))
    elif output_format == "md":
        output = _as_text_writer(output)
        exporter = MarkdownExporter()
        exporter.export(results, output)
    elif output_format == "html":
        output = _as_text_writer(output)
        exporter = HTMLExporter()
        exporter.export(results, output)
    elif output_format == "yaml":
        output = _as_text_writer(output)
        output.write(dump_minimal_yaml(results))
    elif output_format == "turtle":
        output = _as_text_writer(output)
        exporter = RDFExporter()
        exporter.export(results, output, knowledge_engine.schemaview)
    elif output_format == "owl":
        output = _as_text_writer(output)
        exporter = OWLExporter()
        exporter.export(results, output, knowledge_engine.schemaview)
    else:
        output = _as_text_writer(output)
        output.write(dump_minimal_yaml(results))


inputfile_option = click.option("-i", "--inputfile", help="Path to a file containing input text.")
template_option = click.option("-t", "--template", required=True, help="Template to use.")
target_class_option = click.option(
    "-T", "--target-class", help="Target class (if not already root)."
)
interactive_option = click.option(
    "--interactive/--no-interactive",
    default=False,
    show_default=True,
    help="Interactive mode - rather than call the LLM API it will prompt you do this.",
)
model_option = click.option("-m", "--model", help="Engine to use, e.g. text-davinci-003.")
prompt_template_option = click.option(
    "--prompt-template", help="Path to a file containing the prompt."
)
recurse_option = click.option(
    "--recurse/--no-recurse", default=True, show_default=True, help="Recursively parse structures."
)
output_option_wb = click.option(
    "-o", "--output", type=click.File(mode="wb"), default=sys.stdout, help="Output file."
)
output_option_txt = click.option(
    "-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file."
)
output_format_options = click.option(
    "-O",
    "--output-format",
    type=click.Choice(["json", "yaml", "pickle", "md", "html", "owl", "turtle"]),
    default="yaml",
    help="Output format.",
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.option("--cache-db", help="Path to sqlite database to cache prompt-completion results")
@click.option(
    "--skip-annotator", multiple=True, help="Skip annotator (e.g. --skip-annotator gilda)"
)
@click.version_option(__version__)
def main(verbose: int, quiet: bool, cache_db: str, skip_annotator):
    """CLI for oak-ai.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    logger = logging.getLogger()
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)
    logger.info(f"Logger {logger.name} set to level {logger.level}")
    if cache_db:
        settings.cache_db = cache_db
    if skip_annotator:
        settings.skip_annotators = list(skip_annotator)


@main.command()
@inputfile_option
@template_option
@target_class_option
@model_option
@recurse_option
@output_option_wb
@click.option("--dictionary")
@output_format_options
@click.option("--auto-prefix", default="AUTO", help="Prefix to use for auto-generated classes.")
@click.option(
    "--set-slot-value",
    "-S",
    multiple=True,
    help="Set slot value, e.g. --set-slot-value has_participant=protein",
)
@click.argument("input", required=False)
def extract(
    inputfile,
    template,
    target_class,
    dictionary,
    input,
    output,
    output_format,
    set_slot_value,
    **kwargs,
):
    """Extract knowledge from text guided by schema, using SPIRES engine.

    Example:

        ontogpt extract -t gocam.GoCamAnnotations -i gocam-27929086.txt

    The input argument must be either a file path or a string.
    Use the -i/--input-file option followed by the path to the input file if using the former.
    Otherwise, the input is assumed to be a string to be read as input.

    You can also use fragments of existing schemas, use the --target-class option (-T) to
    specify an alternative Container/root class.

    Example:

        ontogpt -extract -t gocam.GoCamAnnotations -T GeneOrganismRelationship "the mouse Shh gene"

    """
    logging.info(f"Creating for {template}")
    ke = SPIRESEngine(template, **kwargs)
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db
    if settings.skip_annotators:
        ke.client.skip_annotators = settings.skip_annotators
    if dictionary:
        ke.load_dictionary(dictionary)
    if inputfile and Path(inputfile).exists():
        text = open(inputfile, "r").read()
    elif inputfile and not Path(inputfile).exists():
        raise FileNotFoundError(f"Cannot find input file {inputfile}")
    elif input:
        text = input
    elif not input or input == "-":
        text = sys.stdin.read()
    logging.info(f"Input text: {text}")
    if target_class:
        target_class_def = ke.schemaview.get_class(target_class)
    else:
        target_class_def = None
    results = ke.extract_from_text(text, target_class_def)
    if set_slot_value:
        for slot_value in set_slot_value:
            slot, value = slot_value.split("=")
            setattr(results.extracted_object, slot, value)
    write_extraction(results, output, output_format, ke)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@click.argument("pmid")
def pubmed_extract(pmid, template, output, output_format, **kwargs):
    """Extract knowledge from a pubmed ID."""
    logging.info(f"Creating for {template}")
    pmc = PubmedClient()
    text = pmc.text(pmid)
    ke = SPIRESEngine(template, **kwargs)
    logging.debug(f"Input text: {text}")
    results = ke.extract_from_text(text)
    write_extraction(results, output, output_format)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@click.option("--auto-prefix", default="AUTO", help="Prefix to use for auto-generated classes.")
@click.argument("article")
def wikipedia_extract(article, template, output, output_format, **kwargs):
    """Extract knowledge from a wikipedia page."""
    logging.info(f"Creating for {template} => {article}")
    client = WikipediaClient()
    text = client.text(article)
    ke = SPIRESEngine(template, **kwargs)
    logging.debug(f"Input text: {text}")
    results = ke.extract_from_text(text)
    write_extraction(results, output, output_format, ke)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@click.option(
    "--keyword",
    "-k",
    multiple=True,
    help="Keyword to search for (e.g. --keyword therapy). Also obtained from schema",
)
@click.argument("topic")
def wikipedia_search(topic, keyword, template, output, output_format, **kwargs):
    """Extract knowledge from a wikipedia page."""
    logging.info(f"Creating for {template} => {topic}")
    client = WikipediaClient()
    keywords = list(keyword) if keyword else []
    logging.info(f"KW={keywords}")
    ke = SPIRESEngine(template, **kwargs)
    keywords.extend(ke.schemaview.schema.keywords)
    search_term = f"{topic + ' ' + ' '.join(keywords)}"
    print(f"Searching for {search_term}")
    search_results = client.search_wikipedia_articles(search_term)
    for _index, result in enumerate(search_results, start=1):
        title = result["title"]
        text = client.text(title)
        logging.debug(f"Input text: {text}")
        if len(text) > 4000:
            # TODO
            text = text[:4000]
        results = ke.extract_from_text(text)
        write_extraction(results, output, output_format)
        break


@main.command()
@template_option
@model_option
@recurse_option
@output_option_txt
@output_format_options
@click.argument("pmcid")
def pmc_extract(pmcid, template, output, output_format, **kwargs):
    """Extract knowledge from PMC (TODO)."""
    logging.info(f"Creating for {template}")
    pmc = PubmedClient()
    ec = pmc.entrez_client
    paset = ec.efetch(db="pmc", id=pmcid)
    from lxml import etree  # noqa

    for pa in paset:
        pa._xml_root
        print(etree.tostring(pa._xml_root, pretty_print=True))


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@click.option(
    "--keyword",
    "-k",
    multiple=True,
    help="Keyword to search for (e.g. --keyword therapy). Also obtained from schema",
)
@click.argument("term_tokens", nargs=-1)
def search_and_extract(term_tokens, keyword, template, output, output_format, **kwargs):
    """Search for relevant literature and extracts knowledge from it."""
    term = " ".join(term_tokens)
    logging.info(f"Creating for {template}; search={term} kw={keyword}")
    ke = SPIRESEngine(template, **kwargs)
    logging.info(f"Creating PubMed client for {template}; search={term}")
    pmc = PubmedClient()
    logging.info("Got client")
    keywords = list(keyword) if keyword else []
    logging.info(f"KW={keywords}")
    keywords.extend(ke.schemaview.schema.keywords)
    logging.info(f"Keywords={keywords}")
    if not keywords:
        raise ValueError("No keywords specified; use --keyword or annotate schema with keywords")
    pmids = list(pmc.search(term, keywords))
    logging.info(f"PMIDs={pmids}")
    pmid = pmids[0]
    logging.info(f"PMID={pmid}")
    text = pmc.text(pmid)
    logging.info(f"Input text: {text}")
    results = ke.extract_from_text(text)
    write_extraction(results, output, output_format)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@click.argument("url")
def web_extract(template, url, output, output_format, **kwargs):
    """Extract knowledge from web page."""
    logging.info(f"Creating for {template}")
    web_client = SoupClient()
    text = web_client.text(url)
    print(f"## Text: \n\n{text}")
    ke = SPIRESEngine(template, **kwargs)
    logging.debug(f"Input text: {text}")
    results = ke.extract_from_text(text)
    write_extraction(results, output, output_format)


@main.command()
@output_option_wb
@click.option("--dictionary")
@output_format_options
@click.option(
    "--recipes-urls-file",
    "-R",
    help="File with URLs to recipes to use for extraction",
)
@click.option("--auto-prefix", default="AUTO", help="Prefix to use for auto-generated classes.")
@click.argument("url")
def recipe_extract(url, recipes_urls_file, dictionary, output, output_format, **kwargs):
    """Extract from recipe on the web."""
    from recipe_scrapers import scrape_me

    if recipes_urls_file:
        with open(recipes_urls_file, "r") as f:
            urls = [line.strip() for line in f.readlines() if url in line]
            if len(urls) != 1:
                raise ValueError(f"Found {len(urls)} URLs in {recipes_urls_file}")
            url = urls[0]
    scraper = scrape_me(url)
    template = "recipe"
    logging.info(f"Creating for {template}")
    ke = SPIRESEngine(template, **kwargs)
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db
    if settings.skip_annotators:
        ke.client.skip_annotators = settings.skip_annotators
    if dictionary:
        ke.load_dictionary(dictionary)
    ingredients = "\n".join(scraper.ingredients())
    instructions = "\n".join(scraper.instructions_list())
    text = f"""
    Recipe: {scraper.title()}
    Ingredients:\n{ingredients}
    Instructions:\n{instructions}
    """
    logging.info(f"Input text: {text}")
    results = ke.extract_from_text(text)
    results.extracted_object.url = url
    write_extraction(results, output, output_format, ke)


@main.command()
@output_option_wb
@output_format_options
@click.argument("input")
def convert(input, output, output_format, **kwargs):
    """Convert output format."""
    template = "recipe"
    logging.info(f"Creating for {template}")
    ke = SPIRESEngine(template, **kwargs)
    cls = ke.template_pyclass
    with open(input, "r") as f:
        data = yaml.safe_load(f)
    obj = cls(**data["extracted_object"])
    results = ExtractionResult(extracted_object=obj)
    write_extraction(results, output, output_format, ke)


@main.command()
@output_option_txt
@output_format_options
@click.option(
    "-C", "--context", required=True, help="domain e.g. anatomy, industry, health-related"
)
@click.argument("term")
def synonyms(term, context, output, output_format, **kwargs):
    """Extract synonyms."""
    logging.info(f"Creating for {term}")
    ke = SynonymEngine()
    out = str(ke.synonyms(term, context))
    output.write(out)


@main.command()
@output_option_txt
@output_format_options
@click.option(
    "--annotation-path",
    "-A",
    required=True,
)
@click.argument("term")
def create_gene_set(term, output, output_format, annotation_path, **kwargs):
    """Create a gene set."""
    logging.info(f"Creating for {term}")
    evaluator = EvalEnrichment()
    evaluator.load_annotations(annotation_path)
    gene_set = evaluator.create_gene_set_from_term(term)
    print(yaml.dump(gene_set.dict(), sort_keys=False))


@main.command()
@output_option_txt
@output_format_options
@click.option("--fill/--no-fill", default=False)
@click.option(
    "--input-file",
    "-U",
    help="File with gene IDs to enrich (if not passed as arguments)",
)
def convert_geneset(input_file, output, output_format, fill, **kwargs):
    """Convert gene set to YAML."""
    gene_set = parse_gene_set(input_file)
    if fill:
        fill_missing_gene_set_values(gene_set)
    output.write(dump_minimal_yaml(gene_set.dict()))


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option(
    "--resolver", "-r", help="OAK selector for the gene ID resolver. E.g. sqlite:obo:hgnc"
)
@click.option(
    "-C",
    "--context",
    help="domain e.g. anatomy, industry, health-related (NOT IMPLEMENTED - currently gene only)",
)
@click.option(
    "--strict/--no-strict",
    default=True,
    show_default=True,
    help="If set, there must be a unique mappings from labels to IDs",
)
@click.option(
    "--show-prompt/--no-show-prompt",
    default=True,
    show_default=True,
    help="If set, show prompt passed to model",
)
@click.option(
    "--input-file",
    "-U",
    help="File with gene IDs to enrich (if not passed as arguments)",
)
@click.option(
    "--randomize-gene-descriptions-using-file",
    help="FOR EVALUATION ONLY. swap out gene descriptions with genes from this gene set filefile",
)
@click.option(
    "--ontological-synopsis/--no-ontological-synopsis",
    default=True,
    show_default=True,
    help="If set, use automated rather than manual gene descriptions",
)
@click.option(
    "--combined-synopsis/--no-combined-synopsis",
    default=False,
    show_default=True,
    help="If set, both gene descriptions",
)
@click.option(
    "--end-marker",
    help="For testing minor variants of prompts",
)
@click.option(
    "--annotations/--no-annotations",
    default=True,
    show_default=True,
    help="If set, include annotations in the prompt",
)
@prompt_template_option
@interactive_option
@click.argument("genes", nargs=-1)
def enrichment(
    genes,
    context,
    input_file,
    resolver,
    output,
    model,
    show_prompt,
    interactive,
    end_marker,
    output_format,
    randomize_gene_descriptions_using_file,
    **kwargs,
):
    """Gene class enrichment.

    Algorithm:

    1. Map gene symbols to IDs using the resolver (unless IDs specified)
    2. Fetch gene descriptions using Alliance API
    3. Create a prompt using descriptions

    Limitations:

    It is very easy to exceed the max token length; resolved in GPT-4?

    Usage:

        ontogpt enrichment -r sqlite:obo:hgnc -U tests/input/genesets/dopamine.yaml

    Usage:

        ontogpt enrichment -r sqlite:obo:hgnc -U tests/input/genesets/dopamine.yaml

    """
    if not genes and not input_file:
        raise ValueError("Either genes or input file must be passed")
    if genes:
        gene_set = GeneSet(name="TEMP", gene_symbols=genes)
    if input_file:
        if genes:
            raise ValueError("Either genes or input file must be passed")
        gene_set = parse_gene_set(input_file)
    if not gene_set:
        raise ValueError("No genes passed")
    ke = create_engine(None, EnrichmentEngine, model=model)
    if end_marker:
        ke.end_marker = end_marker
    if interactive:
        ke.client.interactive = True
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db
    if not isinstance(ke, EnrichmentEngine):
        raise ValueError(f"Expected EnrichmentEngine, got {type(ke)}")
    if resolver:
        ke.add_resolver(resolver)
    if randomize_gene_descriptions_using_file:
        print("WARNING!! Randomly spiking gene descriptions")
        spike_gene_set = parse_gene_set(randomize_gene_descriptions_using_file)
        aliases = {}
        if not spike_gene_set.gene_symbols:
            raise ValueError("No gene symbols for spike set")
        syms = copy(gene_set.gene_symbols)
        if len(spike_gene_set.gene_symbols) < len(gene_set.gene_symbols):
            raise ValueError("Not enough genes in spike set")
        for sym in spike_gene_set.gene_symbols:
            if not syms:
                break
            aliases[sym] = syms.pop()
        results = ke.summarize(
            spike_gene_set, normalize=resolver is not None, gene_aliases=aliases, **kwargs
        )
    else:
        results = ke.summarize(gene_set, normalize=resolver is not None, **kwargs)
    if results.truncation_factor is not None and results.truncation_factor < 1.0:
        logging.warning(f"Text was truncated; factor = {results.truncation_factor}")
    output = _as_text_writer(output)
    if show_prompt:
        print(results.prompt)
    output.write(dump_minimal_yaml(results))


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option(
    "-C",
    "--context",
    help="domain e.g. anatomy, industry, health-related (NOT IMPLEMENTED - currently gene only)",
)
@click.argument("text", nargs=-1)
def embed(text, context, output, model, output_format, **kwargs):
    """Embed text."""
    if not text:
        raise ValueError("Text must be passed")
    if model is None:
        model = "text-embedding-ada-002"
    client = OpenAIClient(model=model)
    resp = client.embeddings(text)
    print(resp)


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option(
    "-C",
    "--context",
    help="domain e.g. anatomy, industry, health-related (NOT IMPLEMENTED - currently gene only)",
)
@click.argument("text", nargs=-1)
def text_similarity(text, context, output, model, output_format, **kwargs):
    """Embed text."""
    if not text:
        raise ValueError("Text must be passed")
    text = list(text)
    if "@" not in text:
        raise ValueError("Text must contain @")
    ix = text.index("@")
    text1 = " ".join(text[:ix])
    text2 = " ".join(text[ix + 1 :])
    print(text1)
    print(text2)
    if model is None:
        model = "text-embedding-ada-002"
    client = OpenAIClient(model=model)
    sim = client.similarity(text1, text2, model=model)
    print(sim)


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option(
    "-C",
    "--context",
    help="domain e.g. anatomy, industry, health-related (NOT IMPLEMENTED - currently gene only)",
)
@click.argument("text", nargs=-1)
def text_distance(text, context, output, model, output_format, **kwargs):
    """Embed text, calculate euclidian distance between embeddings."""
    if not text:
        raise ValueError("Text must be passed")
    text = list(text)
    if "@" not in text:
        raise ValueError("Text must contain @")
    ix = text.index("@")
    text1 = " ".join(text[:ix])
    text2 = " ".join(text[ix + 1 :])
    print(text1)
    print(text2)
    if model is None:
        model = "text-embedding-ada-002"
    client = OpenAIClient(model=model)
    sim = client.euclidian_distance(text1, text2, model=model)
    print(sim)


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option("--ontology", "-r", help="Ontology to use")
@click.option(
    "--definitions/--no-definitions",
    default=True,
    show_default=True,
    help="Include text definitions in the text to embed",
)
@click.option(
    "--parents/--no-parents",
    default=True,
    show_default=True,
    help="Include is-a parent terms in the text to embed",
)
@click.option(
    "--ancestors/--no-ancestors",
    default=True,
    show_default=True,
    help="Include all ancestors in the text to embed",
)
@click.option(
    "--logical-definitions/--no-logical-definitions",
    default=True,
    show_default=True,
    help="Include logical definitions in the text to embed",
)
@click.option(
    "--autolabel/--no-autolabel",
    default=True,
    show_default=True,
    help="Add subj/obj labels to report objects",
)
@click.option(
    "--synonyms/--no-synonyms",
    default=True,
    show_default=True,
    help="Include synonyms in the text to embed",
)
@click.argument("terms", nargs=-1)
def entity_similarity(terms, ontology, output, model, output_format, **kwargs):
    """Embed text.

    Uses ada by default, currently: $0.0004 / 1K tokens
    """
    if not terms:
        raise ValueError("terms must be passed")
    terms = list(terms)
    if "@" not in terms:
        logging.info("No @ found, assuming all by all")
        terms1 = list(terms)
        terms2 = list(terms)
    else:
        ix = terms.index("@")
        terms1 = terms[:ix]
        terms2 = terms[ix + 1 :]
    adapter = get_adapter(ontology)
    entities1 = list(query_terms_iterator(terms1, adapter))
    entities2 = list(query_terms_iterator(terms2, adapter))

    engine = SimilarityEngine(model=model, adapter=adapter, **kwargs)
    writer = StreamingCsvWriter(output, heterogeneous_keys=False)

    for e1 in entities1:
        sims = engine.search(e1, entities2)
        for sim in sims:
            writer.emit(sim)


@main.command()
@output_option_txt
@click.option(
    "--strict/--no-strict",
    default=True,
    show_default=True,
    help="If set, there must be a unique mappings from labels to IDs",
)
@click.option(
    "--input-file",
    "-U",
    help="File with gene IDs to enrich (if not passed as arguments)",
)
@click.option(
    "--ontological-synopsis/--no-ontological-synopsis",
    default=True,
    show_default=True,
    help="If set, use automated rather than manual gene descriptions",
)
@click.option(
    "--combined-synopsis/--no-combined-synopsis",
    default=False,
    show_default=True,
    help="If set, both gene descriptions",
)
@click.option(
    "--annotations/--no-annotations",
    default=True,
    show_default=True,
    help="If set, include annotations in the prompt",
)
@click.option(
    "--number-to-drop",
    "-n",
    type=click.types.INT,
    default=1,
    help="Max number of genes to drop",
)
# @click.option(
#    "--randomize-gene-descriptions/--no-randomize-gene-descriptions",
#    help="DO NOT USE EXCEPT FOR EVALUATION PUPOSES."
# )
@click.option(
    "--annotations-path",
    "-A",
    help="Path to annotations",
)
@click.argument("genes", nargs=-1)
def eval_enrichment(genes, input_file, number_to_drop, annotations_path, output, **kwargs):
    """Run enrichment using multiple methods."""
    if not genes and not input_file:
        raise ValueError("Either genes or input file must be passed")
    if genes:
        gene_set = GeneSet(name="TEMP", gene_symbols=genes)
    if input_file:
        if genes:
            raise ValueError("Either genes or input file must be passed")
        gene_set = parse_gene_set(input_file)
    if not gene_set:
        raise ValueError("No genes passed")
    fill_missing_gene_set_values(gene_set)
    if not annotations_path:
        if not _is_human(gene_set):
            raise ValueError("No annotations path passed")
        annotations_path = "tests/input/genes2go.tsv.gz"
    eval_engine = EvalEnrichment()
    eval_engine.load_annotations(annotations_path)
    comps = eval_engine.evaluate_methods_on_gene_set(gene_set, n=number_to_drop, **kwargs)
    output.write(dump_minimal_yaml(comps))


@main.command()
@model_option
@recurse_option
@output_option_txt
@output_format_options
@click.option(
    "--num-tests",
    type=click.INT,
    default=5,
    show_default=True,
    help="number of iterations to cycle through.",
)
@click.argument("evaluator")
def eval(evaluator, num_tests, output, output_format, **kwargs):
    """Evaluate an extractor."""
    logging.info(f"Creating for {evaluator}")
    evaluator = create_evaluator(evaluator)
    evaluator.num_tests = num_tests
    eos = evaluator.eval()
    output.write(dump_minimal_yaml(eos, minimize=False))


@main.command()
@template_option
@model_option
@click.option("-E", "--examples", type=click.File("r"), help="File of example objects.")
@recurse_option
@output_option_wb
@output_format_options
@click.argument("object")
def fill(template, object: str, examples, output, output_format, **kwargs):
    """Fill in missing values."""
    logging.info(f"Creating for {template}")
    ke = SPIRESEngine(template, **kwargs)
    object = yaml.safe_load(object)
    logging.info(f"Object to fill =  {object}")
    logging.info(f"Loading {examples}")
    examples = yaml.safe_load(examples)
    logging.debug(f"Input object: {object}")
    results = ke.generalize(object, examples)
    output.write(yaml.dump(results.dict()))


@main.command()
def models(**kwargs):
    """Prompt completion."""
    ai = OpenAIClient()
    for model in openai.Model.list():
        print(model)


@main.command()
@model_option
@output_option_txt
@output_format_options
@click.argument("input")
def complete(input, output, output_format, **kwargs):
    """Prompt completion."""
    ai = OpenAIClient()
    text = open(input).read()
    payload = ai.complete(text)
    print(payload)


@main.command()
@template_option
@click.option("--input", "-i", type=click.File("r"), default=sys.stdin, help="Input file")
def parse(template, input):
    """Parse openai results."""
    logging.info(f"Creating for {template}")
    ke = SPIRESEngine(template)
    text = input.read()
    logging.debug(f"Input text: {text}")
    # ke.annotator = BioPortalImplementation()
    results = ke.parse_completion_payload(text)
    print(yaml.dump(results))


@main.command()
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@output_format_options
@model_option
@click.option("-m", "match", help="Match string to use for filtering.")
@click.option("-D", "database", help="Path to sqlite database.")
def dump_completions(engine, match, database, output, output_format):
    """Dump cached completions."""
    logging.info(f"Creating for {engine}")
    client = OpenAIClient()
    if database:
        client.cache_db_path = database
    if output_format == "jsonl":
        writer = jsonlines.Writer(output)
        for engine, prompt, completion in client.cached_completions(match):
            writer.write(dict(engine=engine, prompt=prompt, completion=completion))
    elif output_format == "yaml":
        for engine, prompt, completion in client.cached_completions(match):
            output.write(
                dump_minimal_yaml(dict(engine=engine, prompt=prompt, completion=completion))
            )
    else:
        output.write("# Cached Completions:\n")
        for engine, prompt, completion in client.cached_completions(match):
            output.write("## Entry\n")
            output.write(f"### Engine: {engine}\n")
            output.write(f"### Prompt:\n\n {prompt}\n\n")
            output.write(f"### Completion:\n\n {completion}\n\n")


@main.command()
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@click.argument("input", type=click.File("r"))
def convert_examples(input, output):
    """Convert training examples from YAML."""
    logging.info(f"Creating examples for {input}")
    example_doc = yaml.safe_load(input)
    writer = jsonlines.Writer(output)
    for example in example_doc["examples"]:
        prompt = example["prompt"]
        completion = yaml.dump(example["completion"], sort_keys=False)
        writer.write(dict(prompt=prompt, completion=completion))


@main.command()
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@click.option("-i", "--input", help="Input ontology.")
@click.option("-c", "--context", help="Context.")
@click.option(
    "--num-iterations",
    type=click.INT,
    default=5,
    show_default=True,
    help="number of iterations to cycle through.",
)
@click.argument("terms", nargs=-1)
def halo(input, context, terms, output, **kwargs):
    """Run HALO over inputs."""
    engine = HALOEngine()
    engine.seed_from_file(input)
    if context is None:
        context = engine.ontology.elements[0].context
    engine.fixed_slot_values = {"context": context}
    engine.hallucinate(terms, **kwargs)
    output.write(dump_minimal_yaml(engine.ontology))


@main.command()
@output_option_wb
@output_format_options
@click.option(
    "-d",
    "--description",
    help="domain e.g. anatomy, industry, health-related (NOT IMPLEMENTED - currently gene only)",
)
@click.option(
    "--sections", multiple=True, help="sections to include e.g. medications, vital signs, etc."
)
def clinical_notes(
    description,
    sections,
    output,
    output_format,
    **kwargs,
):
    """Create mock clinical notes.

    Example:

        ontogpt clinical-notes -d "middle-aged female patient with diabetes"
        ontogpt clinical-notes --description "middle-aged female patient with diabetes"\
         --sections medications --sections "vital signs"

    """
    c = OpenAIClient()
    prompt = "create mock clinical notes for a patient like this: " + description
    if sections:
        prompt += " including sections: " + ", ".join(sections)
    results = c.complete(prompt)
    print(results)
    output.write(results)


@main.command()
def list_templates():
    """List the templates."""
    print("TODO")


if __name__ == "__main__":
    main()



"""
CTD evaluation.

Note: the CTD training set has some oddities

E.g. scopolamine induces drug overdose

Biocreativ results here:

https://biocreative.bioinformatics.udel.edu/media/store/files/2015/BC5CDR_overview.final.pdf

A total of 18 teams successfully submitted CID results in 46 runs. As
shown in Table 3 (only the best run of each team is included),
the Fscore ranges from 32.01% to 57.03% (team 288) with an average of
43.37%. All teams outperformed the baseline results by the simple abstract-level
co-occurrence method (16.43% in precision, 76.45% in recall and 27.05% in F-score).

"""
import gzip
import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import Any, Dict, Iterable, List, Set, Tuple

import yaml
from bioc import biocxml
from oaklib import BasicOntologyInterface, get_implementation_from_shorthand
from pydantic import BaseModel

from ontogpt.engines.knowledge_engine import chunk_text
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.evaluation.evaluation_engine import SimilarityScore, SPIRESEvaluationEngine
from ontogpt.templates.core import Publication, Triple
from ontogpt.templates.ctd import (
    ChemicalToDiseaseDocument,
    ChemicalToDiseaseRelationship,
    TextWithTriples,
)

THIS_DIR = Path(__file__).parent
DATABASE_DIR = Path(__file__).parent / "database"
TEST_SET_BIOC = DATABASE_DIR / "CDR_TestSet.BioC.xml.gz"
TRAIN_SET_BIOC = DATABASE_DIR / "CDR_TrainSet.BioC.xml.gz"

RMAP = {"CID": "induces"}

logger = logging.getLogger(__name__)


def negated(Triple) -> bool:
    return Triple.qualifier and Triple.qualifier.lower() == "not"


class PredictionRE(BaseModel):
    """A prediction for a relationship extraction task."""

    test_object: TextWithTriples = None
    """source of truth to evaluate against"""

    true_positives: List[Tuple] = None
    num_true_positives: int = None
    false_positives: List[Tuple] = None
    num_false_positives: int = None
    false_negatives: List[Tuple] = None
    num_false_negatives: int = None
    scores: Dict[str, SimilarityScore] = None
    predicted_object: TextWithTriples = None
    named_entities: List[Any] = None

    def calculate_scores(self, labelers: List[BasicOntologyInterface] = None):
        self.scores = {}

        def label(x):
            if labelers:
                for labeler in labelers:
                    lbl = labeler.label(x)
                    if lbl:
                        return f"{x} {lbl}"
            return x

        def all_objects(dm: TextWithTriples):
            return list(
                set(link.subject for link in dm.triples if not negated(link))
                | set(link.object for link in dm.triples if not negated(link))
            )

        def pairs(dm: TextWithTriples) -> Set:
            return set(
                (label(link.subject), label(link.object))
                for link in dm.triples
                if not negated(link)
            )

        self.scores["similarity"] = SimilarityScore.from_set(
            all_objects(self.test_object),
            all_objects(self.predicted_object),
            labelers=labelers,
        )
        pred_pairs = pairs(self.predicted_object)
        test_pairs = pairs(self.test_object)
        self.true_positives = list(pred_pairs.intersection(test_pairs))
        self.false_positives = list(pred_pairs.difference(test_pairs))
        self.false_negatives = list(test_pairs.difference(pred_pairs))
        self.num_false_negatives = len(self.false_negatives)
        self.num_false_positives = len(self.false_positives)
        self.num_true_positives = len(self.true_positives)


class EvaluationObjectSetRE(BaseModel):
    """A result of predicting relationextractions."""

    precision: float = None
    recall: float = None
    f1: float = None

    training: List[TextWithTriples] = None
    predictions: List[PredictionRE] = None
    test: List[TextWithTriples] = None


@dataclass
class EvalCTD(SPIRESEvaluationEngine):
    # ontology: OboGraphInterface = None
    subject_prefix = "MESH"
    object_prefix = "MESH"

    def __post_init__(self):
        self.extractor = SPIRESEngine("ctd.ChemicalToDiseaseDocument")
        # synonyms are derived entirely from training set
        self.extractor.load_dictionary(DATABASE_DIR / "synonyms.yaml")

    def load_test_cases(self) -> Iterable[ChemicalToDiseaseDocument]:
        return self.load_cases(TEST_SET_BIOC)

    def load_training_cases(self) -> Iterable[ChemicalToDiseaseDocument]:
        return self.load_cases(TRAIN_SET_BIOC)

    def load_cases(self, path: Path) -> Iterable[ChemicalToDiseaseDocument]:
        logger.info(f"Loading {path}")
        with gzip.open(str(path), "rb") as f:
            collection = biocxml.load(f)
            triples_by_text = defaultdict(list)
            for document in collection.documents:
                doc = {}
                for p in document.passages:
                    doc[p.infons["type"]] = p.text
                title = doc["title"]
                abstract = doc["abstract"]
                # text = f"Title: {title} Abstract: {abstract}"
                for r in document.relations:
                    i = r.infons
                    t = Triple(
                        subject=f"{self.subject_prefix}:{i['Chemical']}",
                        predicate=RMAP[i["relation"]],
                        object=f"{self.object_prefix}:{i['Disease']}",
                    )
                    triples_by_text[(title, abstract)].append(t)
        for (title, abstract), triples in triples_by_text.items():
            pub = Publication(title=title, abstract=abstract)
            logger.debug(f"Triples: {len(triples)} for Title: {title} Abstract: {abstract}")
            yield ChemicalToDiseaseDocument(publication=pub, triples=triples)

    def create_training_set(self, num=100):
        ke = self.extractor
        docs = list(self.load_test_cases())
        shuffle(docs)
        for doc in docs[0:num]:
            text = doc.text
            prompt = ke.get_completion_prompt(None, text)
            completion = ke.serialize_object(m)
            yield dict(prompt=prompt, completion=completion)

    def eval(self) -> EvaluationObjectSetRE:
        """Evaluate the ability to extract relations."""
        labeler = get_implementation_from_shorthand("sqlite:obo:mesh")
        num_test = self.num_tests
        ke = self.extractor
        docs = list(self.load_test_cases())
        shuffle(docs)
        eos = EvaluationObjectSetRE(
            test=docs[:num_test],
            training=[],
            predictions=[],
        )
        n = 1
        for doc in eos.test:
            logger.info(f"Iteration {n} of {num_test}")
            n += 1
            logger.info(doc)
            text = f"Title: {doc.publication.title} Abstract: {doc.publication.abstract}"
            predicted_obj = None
            named_entities = []
            for chunked_text in chunk_text(text):
                extraction = ke.extract_from_text(chunked_text)
                logger.info(
                    f"{len(extraction.extracted_object.triples)}\
                        triples from window: {chunked_text}"
                )
                if not predicted_obj:
                    predicted_obj = extraction.extracted_object
                else:
                    predicted_obj.triples.extend(extraction.extracted_object.triples)
                    logger.info(f"{len(predicted_obj.triples)} total triples, after concatenation")
                    logger.debug(f"concatenated triples: {predicted_obj.triples}")
                named_entities.extend(extraction.named_entities)

            # title_extraction = ke.extract_from_text(doc.publication.title)
            # logger.info(f"{len(title_extraction.extracted_object.triples)}\
            # triples from: Title {doc.publication.title}")
            # abstract_extraction = ke.extract_from_text(doc.publication.abstract)
            # logger.info(f"{len(abstract_extraction.extracted_object.triples)}\
            # triples from: Abstract {doc.publication.abstract}")
            # ke.merge_resultsets([results, results2])
            # predicted_obj = title_extraction.extracted_object
            # predicted_obj.triples.extend(abstract_extraction.extracted_object.triples)
            # logger.info(f"{len(predicted_obj.triples)} total triples, after concatenation")
            # logger.debug(f"concatenated triples: {predicted_obj.triples}")

            def included(t: ChemicalToDiseaseRelationship):
                return (
                    t
                    and t.subject
                    and t.object
                    and t.subject.startswith("MESH:")
                    and t.object.startswith("MESH:")
                    and t.predicate.lower() == "induces"
                )

            predicted_obj.triples = [t for t in predicted_obj.triples if included(t)]
            logger.info(
                f"{len(predicted_obj.triples)} filtered triples (CID only, between MESH only)"
            )
            pred = PredictionRE(predicted_object=predicted_obj, test_object=doc)
            pred.named_entities = named_entities
            logger.info("PRED")
            logger.info(yaml.dump(pred.dict()))
            logger.info("Calc scores")
            pred.calculate_scores(labelers=[labeler])
            logger.info(yaml.dump(pred.dict()))
            eos.predictions.append(pred)
        self.calc_stats(eos)
        return eos

    def calc_stats(self, eos: EvaluationObjectSetRE):
        num_true_positives = sum(p.num_true_positives for p in eos.predictions)
        num_false_positives = sum(p.num_false_positives for p in eos.predictions)
        num_false_negatives = sum(p.num_false_negatives for p in eos.predictions)
        if num_true_positives + num_false_positives == 0:
            logger.warning("No true positives or false positives")
            return
        eos.precision = num_true_positives / (num_true_positives + num_false_positives)
        eos.recall = num_true_positives / (num_true_positives + num_false_negatives)
        if eos.precision + eos.recall == 0:
            logger.warning("No precision or recall")
            return
        eos.f1 = 2 * (eos.precision * eos.recall) / (eos.precision + eos.recall)



"""Recipe template."""
from __future__ import annotations

from typing import Any, List, Optional

from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class Recipe(ConfiguredBaseModel):
    url: Optional[str] = Field(None)
    label: Optional[str] = Field(None, description="""the name of the recipe""")
    description: Optional[str] = Field(
        None, description="""a brief textual description of the recipe"""
    )
    categories: Optional[List[str]] = Field(
        default_factory=list,
        description="""a semicolon separated list of the categories to which this recipe belongs""",
    )
    ingredients: Optional[List[Ingredient]] = Field(
        default_factory=list,
        description="""a semicolon separated list of the\
            ingredients plus quantities of the recipe""",
    )
    steps: Optional[List[Step]] = Field(
        default_factory=list,
        description="""a semicolon separated list of the\
            individual steps involved in this recipe""",
    )


class ExtractionResult(ConfiguredBaseModel):
    """A result of extracting knowledge on text."""

    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(
        None, description="""The complex objects extracted from the text"""
    )
    named_entities: Optional[List[Any]] = Field(
        default_factory=list, description="""Named entities extracted from the text"""
    )


class NamedEntity(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class FoodType(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class RecipeCategory(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Action(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class UtensilType(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Unit(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class CompoundExpression(ConfiguredBaseModel):
    pass


class Ingredient(CompoundExpression):
    food_item: Optional[FoodItem] = Field(None, description="""the food item""")
    amount: Optional[Quantity] = Field(
        None, description="""the quantity of the ingredient, e.g. 2 lbs"""
    )


class Quantity(CompoundExpression):
    value: Optional[str] = Field(None, description="""the value of the quantity""")
    unit: Optional[str] = Field(
        None, description="""the unit of the quantity, e.g. grams, cups, etc."""
    )


class Step(CompoundExpression):
    action: Optional[str] = Field(
        None, description="""the action taken in this step (e.g. mix, add)"""
    )
    inputs: Optional[List[FoodItem]] = Field(
        default_factory=list,
        description="""a semicolon separated list of the inputs of this step""",
    )
    outputs: Optional[List[FoodItem]] = Field(
        default_factory=list,
        description="""a semicolon separated list of the outputs of this step""",
    )
    utensils: Optional[List[str]] = Field(
        default_factory=list,
        description="""the kitchen utensil used in this step (e.g. pan, bowl)""",
    )


class FoodItem(CompoundExpression):
    food: Optional[str] = Field(None, description="""the food item""")
    state: Optional[str] = Field(
        None, description="""the state of the food item (e.g. chopped, diced)"""
    )


class Triple(CompoundExpression):
    """Abstract parent for Relation Extraction tasks."""

    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(
        None, description="""A qualifier for the statements, e.g. \"NOT\" for negation"""
    )
    subject_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the subject of the\
            statement, e.g. \"high dose\" or \"intravenously administered\"""",
    )
    object_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the object of\
            the statement, e.g. \"severe\" or \"with additional complications\"""",
    )


class TextWithTriples(ConfiguredBaseModel):
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
Recipe.update_forward_refs()
ExtractionResult.update_forward_refs()
NamedEntity.update_forward_refs()
FoodType.update_forward_refs()
RecipeCategory.update_forward_refs()
Action.update_forward_refs()
UtensilType.update_forward_refs()
Unit.update_forward_refs()
CompoundExpression.update_forward_refs()
Ingredient.update_forward_refs()
Quantity.update_forward_refs()
Step.update_forward_refs()
FoodItem.update_forward_refs()
Triple.update_forward_refs()
TextWithTriples.update_forward_refs()
RelationshipType.update_forward_refs()
Publication.update_forward_refs()
AnnotatorResult.update_forward_refs()



"""
Uses code-davinci-002.

Note also that fine-tuning can't be done with code-davinci-002, see:
https://community.openai.com/t/finetuning-code-davinci/23132/2
"""
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import openai
import pydantic
import tiktoken
import yaml
from linkml.utils.schema_fixer import uncamel
from linkml_runtime.utils.formatutils import camelcase
from oaklib.datamodels.obograph import Graph
from oaklib.datamodels.vocabulary import IS_A
from oaklib.interfaces.obograph_interface import OboGraphInterface
from tiktoken import Encoding

from ontogpt.clients import OpenAIClient
from ontogpt.engines.knowledge_engine import FIELD, KnowledgeEngine
from ontogpt.io.yaml_wrapper import dump_minimal_yaml
from ontogpt.templates.halo import Ontology, OntologyElement

this_path = Path(__file__).parent
logger = logging.getLogger(__name__)


ELEMENT_NAME = str

INSTRUCTIONS = """
## Instructions:
## Add an additional element to the YAML below, which is for elements
## in an industrial ontology. Complete as far as possible the following
## fields:
"""


class StructuredPrompt(pydantic.BaseModel):
    header: str = None
    body: str = None
    main_prompt: str = None

    @property
    def text(self) -> str:
        return f"{self.header}\n{self.body}\n{self.main_prompt}"


@dataclass
class HALOEngine(KnowledgeEngine):
    """Engine for Hallucinating Latent Ontologies."""

    engine: str = "code-davinci-002"
    ontology: Ontology = None
    traverse_slots: List[FIELD] = field(
        default_factory=lambda: ["subtypes", "parts", "subclass_of", "part_of"]
    )
    fixed_slot_values: Dict[str, str] = None
    adapter: OboGraphInterface = None
    visited: Set[ELEMENT_NAME] = field(default_factory=lambda: set())
    candidates: List[ELEMENT_NAME] = None
    always_extend: bool = False
    expand_horizon: bool = False

    element_scores: Dict[ELEMENT_NAME, float] = field(default_factory=lambda: {})
    """Ranks each element by estimated informativeness for training."""

    tokenizer_encoding: Encoding = field(default_factory=lambda: tiktoken.get_encoding("gpt2"))

    def __post_init__(self):
        self.template_class = self._get_template_class("halo.OntologyElement")
        self.client = OpenAIClient(model=self.engine)
        self.api_key = self._get_openai_api_key()
        openai.api_key = self.api_key

    def seed(self, seed_ontology: Ontology):
        """Seed the engine with an initial ontology.

        :param seed_ontology:
        :return:
        """
        self.ontology = seed_ontology
        if not self.expand_horizon:
            self.visited = {x.name for x in self.ontology.elements}

    def seed_from_file(self, file_path: str) -> Ontology:
        """Seed the engine with an initial ontology from a file.

        :param file_path:
        :return:
        """
        ontology = Ontology(**yaml.safe_load(open(file_path)))
        self.seed(ontology)
        logger.info(f"Seeded with {len(ontology.elements)} elements")
        return ontology

    def hallucinate(
        self, seed_elements: List[ELEMENT_NAME] = None, num_iterations=10
    ) -> List[OntologyElement]:
        """Run the HALO engine for a given number of iterations.

        Each iteration will expand the initial seed ontology.

        :param num_iterations:
        :return:
        """
        added = []
        if seed_elements:
            for e in seed_elements:
                added.append(self.hallucinate_element(e))
        for i in range(num_iterations):
            logger.info(f"Running HALO iteration {i}")
            elt = self.hallucinate_once()
            if elt:
                added.append(elt)
            else:
                break
        return added

    def hallucinate_once(self) -> Optional[OntologyElement]:
        """Run the HALO engine once.

        Finds a candidate element to expand,
        then runs HALO on that element by generating
        a prompt for it.

        :return:
        """
        candidate_elements = self.get_candidate_elements()
        logger.info(f"Found {len(candidate_elements)} candidate elements")
        logger.info(f"Candidate elements: {candidate_elements}  // visited={self.visited}")
        if not candidate_elements:
            return None
        element = candidate_elements[0]
        logger.info(f"Selected element {element}")
        return self.hallucinate_element(element)

    def get_candidate_elements(self) -> List[ELEMENT_NAME]:
        """Get candidate elements to expand.

        Has side effect of removing candidate cache if visited
        :return:
        """
        if self.candidates is None:
            # TODO: exclude seed set that is not on the horizon
            self.candidates = [x.name for x in self.ontology.elements]
        self.candidates = [c for c in self.candidates if c not in self.visited]
        return self.candidates

    def extend_candidates(self, elements: List[ELEMENT_NAME]) -> List[ELEMENT_NAME]:
        """Extend candidates by all entities in the signature of the specified elements.

        :return:
        """
        visited = self.visited
        if not self.traverse_slots:
            raise ValueError("No slots to traverse")
        for element_name in elements:
            visited.add(element_name)
            logger.info(f"Extending candidates for {element_name}; visited={visited}")
            element = self.get_element(element_name)
            for slot_name in self.traverse_slots:
                refs = getattr(element, slot_name)
                for ref in refs:
                    if ref not in visited:
                        if not self.candidates:
                            self.candidates = []
                        self.candidates.append(ref)
                        logger.info(f" -- Added {ref} to candidates")
        return self.candidates

    def get_element(self, element_name: ELEMENT_NAME) -> Optional[OntologyElement]:
        """Get an element by name.

        :param element_name:
        :return:
        """
        for e in self.ontology.elements:
            if e.name == element_name:
                return e

    def old_get_candidate_elements(self) -> List[ELEMENT_NAME]:
        """Get candidate elements for HALO.

        :return:
        """
        candidate_elements = set()
        visited = self.visited
        if not self.traverse_slots:
            raise ValueError("No slots to traverse")
        for element in self.ontology.elements:
            all_slots_have_refs = True
            for slot_name in self.traverse_slots:
                refs = getattr(element, slot_name)
                if len(refs) == 0 and element.name not in visited:
                    candidate_elements.add(element.name)
                    all_slots_have_refs = False
                for ref in refs:
                    if ref not in visited:
                        candidate_elements.add(ref)
                        visited.add(ref)
            if all_slots_have_refs and not self.always_extend:
                visited.add(element.name)
        return list(candidate_elements)

    def hallucinate_element(self, element: ELEMENT_NAME) -> OntologyElement:
        """Generate an ontology element based on its name.

        :param element: example: LeftDigit1OfHand
        :return:
        """
        logger.info(f"Hallucinating element {element}")
        example_elements = self.get_example_elements(element)
        example_element_names = [x.name for x in example_elements]
        logger.info(f"Found {len(example_elements)} example elements: {example_element_names}")
        # TODO: set bound dynamically
        prompt = self.generate_prompt(element, example_elements[0:10])
        # logger.info(f"Generated prompt: {prompt}")
        payload = self.client.complete(prompt.text)
        objs = self.integrate_payload(prompt, payload)
        logger.info(f"Integrated {len(objs)} objects")
        obj = objs[0]
        self.extend_candidates([obj.name])
        return obj

    def get_example_elements(self, element: ELEMENT_NAME) -> List[OntologyElement]:
        """Get example elements for HALO.

        :param element:
        :return:
        """
        name = uncamel(element)
        toks = set(self.tokenizer_encoding.encode(name))
        score_element_pairs = [(self.get_element_score(e, toks), e) for e in self.ontology.elements]
        score_element_pairs.sort(key=lambda x: x[0], reverse=True)
        logger.info(f"Scores[{element}: {[x for x, _ in score_element_pairs]}")
        logger.info(f"Sorted elements: {score_element_pairs}")
        return [x for _, x in score_element_pairs]

    def get_element_score(self, element: OntologyElement, tokens: Set[int]) -> float:
        """Calculate a score for an element based on how informative it is for few-shot learning.

        :param element:
        :param tokens: tokenized form of the element name
        :return:
        """
        element_name = element.name
        if element_name in self.element_scores:
            score = self.element_scores[element_name]
        else:
            score = 0
            for _, v in element.dict().items():
                if v:
                    score += 1
            self.element_scores[element_name] = score
        element_tokens = set(self.tokenizer_encoding.encode(element_name))
        jaccard = len(tokens.intersection(element_tokens)) / len(tokens.union(element_tokens))
        return score / 100 + jaccard

    def generate_prompt(
        self, seed_element: ELEMENT_NAME, example_elements: List[OntologyElement]
    ) -> StructuredPrompt:
        """Generate a prompt for HALO.

        :param seed_element:
        :param example_elements:
        :return:
        """
        prompt = StructuredPrompt()
        prompt.header = INSTRUCTIONS
        for slot in self.schemaview.class_induced_slots("OntologyElement"):
            desc = slot.description
            if not desc:
                logger.warning(f"No description for slot {slot.name}")
                desc = ""
            prompt.header += f"## {slot.name}: {desc}\n"
        prompt.body = "\n## Examples:\n"
        for element in example_elements:
            prompt.body += dump_minimal_yaml(element)
        stub_object = {
            "name": seed_element,
        }
        for k, v in self.fixed_slot_values.items():
            stub_object[k] = v
        prompt.main_prompt = yaml.dump([stub_object])
        logger.info(
            f"Generated prompt: {len(prompt.text)} = {len(prompt.header)} +\
                {len(prompt.body)} + {len(prompt.main_prompt)}"
        )
        return prompt

    def integrate_payload(
        self, prompt: StructuredPrompt, payload: Dict[str, Any]
    ) -> List[OntologyElement]:
        """Integrate the payload from HALO into the ontology.

        :param payload:
        :param element:
        :return:
        """
        effective_payload = prompt.main_prompt + payload
        # logger.info(f"## EFFECTIVE: {effective_payload}")
        try:
            objs = yaml.safe_load(effective_payload)
        except:
            # codex does not give reliable YAML
            objs = self.parse_what_you_can(effective_payload)
        logger.info(f"## PARSED: {len(objs)}")
        elt = self.integrate_object(objs[0])
        logger.info(f" * INTEGRATED: {elt}")
        return [elt]

    def integrate_object(self, obj: Dict[str, Any], strict=True) -> Optional[OntologyElement]:
        obj = self.repair_dict(obj)
        try:
            elt = OntologyElement(**obj)
        except pydantic.ValidationError as e:
            logger.warning(f"## COULD NOT PARSE: {obj} /// {e}")
            if strict:
                raise e
            return None
        self.add_element(elt)
        return elt

    def repair_dict(self, obj: dict) -> dict:
        slots = self.schemaview.class_slots("OntologyElement")
        nu_obj = {}
        for k, v in obj.items():
            if k not in slots:
                logger.warning(f"Could not find slot {k} in slots")
                continue
            slot = self.schemaview.induced_slot(k, "OntologyElement")
            if slot.multivalued and not isinstance(v, list):
                logger.warning(f"Coercing {v} to list")
                v = [v]
            elif not slot.multivalued and isinstance(v, list):
                logger.warning(f"Coercing {v} len {len(v)} to single value")
                v = v[0]
            nu_obj[k] = v
        return nu_obj

    def old_integrate_payload(self, prompt: StructuredPrompt, payload: Dict[str, Any]):
        """Integrate the payload from HALO into the ontology.

        :param payload:
        :param element:
        :return:
        """
        allowed_slots = self.schemaview.class_slots("OntologyElement")
        effective_payload = prompt.main_prompt + payload
        # logger.info(f"## EFFECTIVE: {effective_payload}")
        try:
            objs = yaml.safe_load(effective_payload)
        except:
            # codex does not give reliable YAML
            objs = self.parse_what_you_can(effective_payload)
        logger.info(f"## PARSED: {len(objs)}")
        added = []
        n = 0
        for obj in objs:
            n += 1
            slots_populated = [k for k, v in obj.items() if v]
            diff = set(slots_populated).difference(allowed_slots)
            if diff:
                logger.info(f"## SKIPPING SLOTS {diff}")
            obj = {k: v for k, v in obj.items() if k in allowed_slots}
            try:
                elt = OntologyElement(**obj)
            except pydantic.ValidationError as e:
                logger.info(f"## COULD NOT PARSE: {obj} /// {e}")
                return added
            logger.info(f"Elt: {elt.name} // {slots_populated} // {obj}")
            if self.add_element(elt):
                logger.info(f" - Added {elt.name}")
                added.append(elt)
            else:
                logger.info(f" - already got {elt.name}")
                if n == 1:
                    logger.error(f"Failed to add first element {elt.name}")
        logger.info(f"Added {len(added)} elements")
        return added

    def parse_what_you_can(self, yaml_str: str) -> List[Dict[str, Any]]:
        """Parse as much of the YAML as possible.

        :param yaml_str:
        :return:
        """
        objs = None
        chunk = ""
        for line in yaml_str.split("\n"):
            chunk += line + "\n"
            try:
                objs = yaml.safe_load(chunk)
            except:
                pass
        if objs is None:
            raise ValueError(f"Could not parse YAML {yaml_str}")
        return objs

    def add_element(self, element: OntologyElement) -> bool:
        """Add an element to the ontology.

        :param obj:
        :return:
        """
        existing = self.get_element(element.name)
        if existing:
            return False
        self.ontology.elements.append(element)
        return True

    def xxextract_seed_ontology(self, seeds: List[str], predicates: List[str]) -> Ontology:
        """Extract an ontology from a given text.

        :param text:
        :return:
        """
        ancestors = list(set(list(self.adapter.ancestors(seeds, predicates, reflexive=True))))
        seed_graph = self.adapter.extract_graph(ancestors, predicates, dangling=False)
        logger.info(len(seed_graph.nodes))
        seed_ontology = self.ontology_from_obograph(seed_graph)
        return seed_ontology

    def xxontology_from_obograph(self, graph: Graph) -> Ontology:
        """Convert an OBO Graph to an Ontology.

        :param graph:
        :return:
        """
        adapter = self.adapter
        ontology = Ontology()
        element_index = {}
        node_to_element_name = {}
        id2slot = {}
        inverses = {}
        for slot in self.schemaview.class_induced_slots(OntologyElement.__name__):
            if slot.inverse:
                inverses[slot.name] = slot.inverse
                inverses[slot.inverse] = slot.name
            if slot.slot_uri:
                id2slot[slot.slot_uri] = slot
        logger.info(list(id2slot.keys()))
        logger.info(inverses)
        for node in graph.nodes:
            meta = node.meta
            if not node.lbl:
                continue
            if not meta:
                # logger.warning(f"Node {node.id} has no meta")
                continue
            element = OntologyElement(
                name=self.node_to_name(node.id, node.lbl),
                synonyms=[synonym.val for synonym in meta.synonyms],
                description=meta.definition.val if meta.definition else None,
            )
            for k, v in self.fixed_slot_values.items():
                setattr(element, k, v)
            element_index[element.name] = element
            node_to_element_name[node.id] = element.name
        for edge in graph.edges:
            if edge.pred == "is_a":
                pred = IS_A
            else:
                try:
                    pred = adapter.uri_to_curie(edge.pred)
                except:
                    pred = edge.pred
            if pred not in id2slot:
                continue
            if edge.sub not in node_to_element_name:
                continue
            if edge.obj not in node_to_element_name:
                continue
            subject = node_to_element_name[edge.sub]
            object = node_to_element_name[edge.obj]
            slot = id2slot[pred]
            getattr(element_index[subject], slot.name).append(object)
            if slot.name in inverses:
                inverse = inverses[slot.name]
                getattr(element_index[object], inverse).append(subject)
        for ldef in adapter.logical_definitions([node.id for node in graph.nodes]):
            if ldef.definedClassId in node_to_element_name:
                element = element_index[node_to_element_name[ldef.definedClassId]]
                if not ldef.genusIds:
                    continue
                if not ldef.restrictions:
                    continue
                genus_elts = [node_to_element_name[g] for g in ldef.genusIds]
                differentia = [
                    f"{adapter.label(r.propertyId)} some {self.node_to_name(r.fillerId)}"
                    for r in ldef.restrictions
                ]
                element.equivalent_to = (
                    f"{' and '.join(genus_elts)} and {' and '.join(differentia)}"
                )
                logger.info(f"Equiv[{element.name}] = {element.equivalent_to}")
        for element in element_index.values():
            ontology.elements.append(element)
        return ontology

    def xxnode_to_name(self, curie: str, label: Optional[str] = None) -> str:
        """Convert a node to a name.

        :param curie:
        :param label:
        :return:
        """
        if label is None:
            label = self.adapter.label(curie)
        if label is None:
            logger.warning(f"Node {curie} has no label")
            label = curie
        return camelcase(label)



"""OpenAI client."""
import ast
import logging
import sqlite3
import sys
from dataclasses import dataclass, field
from pathlib import Path
from time import sleep
from typing import Iterator, Tuple

import numpy as np
import openai
from oaklib.utilities.apikey_manager import get_apikey_value

logger = logging.getLogger(__name__)
NUM_RETRIES = 3


@dataclass
class OpenAIClient:
    # max_tokens: int = field(default_factory=lambda: 3000)
    model: str = field(default_factory=lambda: "gpt-3.5-turbo")
    cache_db_path: str = None
    api_key: str = None
    interactive: bool = None

    def __post_init__(self):
        if not self.api_key:
            self.api_key = get_apikey_value("openai")
        openai.api_key = self.api_key

    def complete(self, prompt, max_tokens=3000, **kwargs) -> str:
        engine = self.model
        logging.info(f"Complete: engine={engine}, prompt={prompt[0:100]}...")
        cur = self.db_connection()
        res = cur.execute("SELECT payload FROM cache WHERE prompt=? AND engine=?", (prompt, engine))
        payload = res.fetchone()
        if payload:
            prompt_peek = str(prompt)[0:80].replace("\n", "\\n")
            logger.info(f"Using cached payload for prompt: {prompt_peek}...")
            return payload[0]
        response = None
        i = 0
        while not response:
            i += 1
            logging.debug(f"Calling OpenAI API (attempt {i})...")
            try:
                if self.interactive:
                    response = self._interactive_completion(prompt, engine, max_tokens, **kwargs)
                elif self._must_use_chat_api():
                    response = openai.ChatCompletion.create(
                        model=engine,
                        messages=[
                            {
                                "role": "user",
                                "content": prompt,
                            },
                        ],
                        max_tokens=max_tokens,
                        **kwargs,
                    )
                else:
                    response = openai.Completion.create(
                        engine=engine,
                        prompt=prompt,
                        max_tokens=max_tokens,
                    )
                break
            except Exception as e:
                logger.error(f"OpenAI API connection error: {e}")
                if i >= NUM_RETRIES:
                    raise e
                sleep_time = 4**i
                logger.info(f"Retrying {i} of {NUM_RETRIES} after {sleep_time} seconds...")
                sleep(sleep_time)

        if self.interactive:
            payload = response
        elif self._must_use_chat_api():
            payload = response["choices"][0]["message"]["content"]
        else:
            payload = response["choices"][0]["text"]
        logger.info(f"Storing payload of len: {len(payload)}")
        cur.execute(
            "INSERT INTO cache (prompt, engine, payload) VALUES (?, ?, ?)",
            (prompt, engine, payload),
        )
        cur.connection.commit()
        return payload

    def db_connection(self):
        if not self.cache_db_path:
            self.cache_db_path = ".openai_cache.db"
        logger.info(f"Caching OpenAI responses to {Path(self.cache_db_path).absolute()}")
        create = not Path(self.cache_db_path).exists()
        con = sqlite3.connect(self.cache_db_path)
        cur = con.cursor()
        if create:
            cur.execute("CREATE TABLE cache (prompt, engine, payload)")
        return cur

    def _interactive_completion(self, prompt: str, engine: str, max_tokens: int = None, **kwargs):
        print("Please use the ChatGPT interface to complete the following prompt:")
        print(f"IMPORTANT: make sure model == {engine}")
        print(f"Note: max_tokens == {max_tokens}")
        print("Prompt:")
        print(prompt)
        print("Paste in the response here, then ctrl-d to continue:")
        response = sys.stdin.read()
        print("OK? (y/n)")
        ok = input()
        if ok == "y":
            print("Thank you! This will now be cached.")
            print("Please be patient for the rest of the process to finish...")
            return response
        else:
            return self._interactive_completion(prompt, engine, max_tokens, **kwargs)

    def cached_completions(
        self, search_term: str = None, engine: str = None
    ) -> Iterator[Tuple[str, str, str]]:
        if search_term:
            search_term = search_term.lower()
        cur = self.db_connection()
        res = cur.execute("SELECT engine, prompt, payload FROM cache")
        for row in res:
            if (
                search_term
                and search_term not in row[1].lower()
                and search_term not in row[2].lower()
            ):
                continue
            if engine and engine != row[0]:
                continue
            yield row

    def _must_use_chat_api(self) -> bool:
        """Return True if the model requires the chat API, False otherwise."""
        if self.model.startswith("text-davinci"):
            return False
        return True

    def embeddings(self, text: str, model: str = None):
        if model is None:
            model = "text-embedding-ada-002"
        cur = self.db_connection()
        try:
            logger.info("creating embeddings cache")
            cur.execute("CREATE TABLE embeddings_cache (text, engine, vector_as_string)")
        except sqlite3.OperationalError:
            logger.info("Embeddings cache table already exists")
            pass
        res = cur.execute(
            "SELECT vector_as_string FROM embeddings_cache WHERE text=? AND engine=?", (text, model)
        )
        payload = res.fetchone()
        if payload:
            logger.info(f"Using cached embeddings for {model} {text[0:80]}...")
            return ast.literal_eval(payload[0])
        logger.info(f"querying OpenAI for {model} {text[0:80]}...")
        response = openai.Embedding.create(
            model=model,
            input=text,
        )
        v = response.data[0]["embedding"]
        logger.info(f"Storing embeddings of len: {len(v)}")
        cur.execute(
            "INSERT INTO embeddings_cache (text, engine, vector_as_string) VALUES (?, ?, ?)",
            (text, model, str(v)),
        )
        cur.connection.commit()
        return v

    def similarity(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        logger.debug(f"similarity: {a1[0:10]}... x {a2[0:10]}... // ({len(a1)} x {len(a2)})")
        return np.dot(a1, a2) / (np.linalg.norm(a1) * np.linalg.norm(a2))

    def euclidian_distance(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        return np.linalg.norm(np.array(a1) - np.array(a2))



from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from linkml_runtime.linkml_model import Decimal
from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class SortFieldEnum(str, Enum):
    ANY = "ANY"
    P_VALUE = "P_VALUE"


class ClassEnrichmentConfiguration(ConfiguredBaseModel):
    """
    configuration for search
    """

    p_value_cutoff: float = Field(None, description="""p-value cutoff for enrichment""")


class ClassEnrichmentResultSet(ConfiguredBaseModel):
    """
    A collection of enrichemt results
    """

    results: Optional[List[ClassEnrichmentResult]] = Field(
        default_factory=list, description="""The enrichment results"""
    )


class ClassEnrichmentResult(ConfiguredBaseModel):
    """
    A single enrichment result
    """

    class_id: str = Field(None, description="""The class id""")
    class_label: Optional[str] = Field(None, description="""The class label""")
    rank: Optional[int] = Field(None, description="""The rank of this result""")
    p_value: float = Field(None, description="""The p-value""")
    p_value_adjusted: Optional[float] = Field(None, description="""The adjusted p-value""")
    false_discovery_rate: Optional[float] = Field(None, description="""The false discovery rate""")
    fold_enrichment: Optional[float] = Field(None, description="""The fold enrichment""")
    probability: Optional[float] = Field(
        None, description="""The probability, as estimated by model-based approaches""", ge=0, le=1
    )
    sample_count: Optional[int] = Field(
        None, description="""The number of entities in the sample with this class"""
    )
    sample_total: Optional[int] = Field(
        None, description="""The total number of entities in the sample"""
    )
    background_count: Optional[int] = Field(None, description="""The background count""")
    background_total: Optional[int] = Field(None, description="""The background total""")
    ancestor_of_more_informative_result: Optional[bool] = Field(
        None, description="""This term is more general than a previously reported result"""
    )
    descendant_of_more_informative_result: Optional[bool] = Field(
        None, description="""This term is more specific than a previously reported result"""
    )


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
ClassEnrichmentConfiguration.update_forward_refs()
ClassEnrichmentResultSet.update_forward_refs()
ClassEnrichmentResult.update_forward_refs()



"""RDF convertor."""
from dataclasses import dataclass
from pathlib import Path
from typing import TextIO, Union

from linkml.generators.pythongen import PythonGenerator
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import rdflib_dumper

from ontogpt.io.exporter import Exporter
from ontogpt.templates.core import ExtractionResult


@dataclass
class RDFExporter(Exporter):
    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO],
        schemaview: SchemaView,
        id_value=None,
    ):
        if isinstance(output, Path):
            output = str(output)
        if isinstance(output, str):
            output = open(str(output), "w", encoding="utf-8")
        element = extraction_output.extracted_object
        cls_name = type(element).__name__
        element_dict_obj = element.dict()
        dc_mod = self._dataclass_model(schemaview)
        dc_cls = dc_mod.__dict__[cls_name]
        dc_obj = dc_cls(**element_dict_obj)
        pm = {"_base": "http://example.org/NOPREFIX/"}
        if "AUTO" not in schemaview.schema.prefixes:
            pm["AUTO"] = "http://example.org/AUTO/"
        dmp = rdflib_dumper.dumps(dc_obj, schemaview=schemaview, prefix_map=pm)
        output.write(dmp)

    def _dataclass_model(self, schemaview: SchemaView):
        schemaview.merge_imports()
        return PythonGenerator(schemaview.schema).compile_module()



"""Clients for accessing different APIs."""
from ontogpt.clients.openai_client import OpenAIClient  # noqa:F401
from ontogpt.clients.pubmed_client import PubmedClient  # noqa:F401
from ontogpt.clients.soup_client import SoupClient  # noqa:F401



"""Evaluation Resolver."""
from typing import Optional, Type, Union

from class_resolver import ClassResolver

from ontogpt.evaluation.ctd.eval_ctd import EvalCTD
from ontogpt.evaluation.evaluation_engine import SPIRESEvaluationEngine

resolver = ClassResolver([EvalCTD], base=SPIRESEvaluationEngine)


def create_evaluator(name: Optional[Union[str, Type]] = None, **kwargs) -> SPIRESEvaluationEngine:
    """Create a knowledge engine."""
    if name is None:
        name = EvalCTD
    if isinstance(name, str):
        return resolver.lookup(name)(**kwargs)
    return name(**kwargs)



from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from linkml_runtime.linkml_model import Decimal
from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class GeneDescription(ConfiguredBaseModel):
    """
    A summarization of an individual gene
    """

    about: Optional[str] = Field(None)
    narrative_summary: Optional[str] = Field(
        None, description="""A free text summary describing the function of the gene"""
    )
    terms: Optional[List[str]] = Field(
        default_factory=list,
        description="""A semicolon separated list of controlled terms that describe the function of the gene""",
    )


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """

    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(
        None, description="""The complex objects extracted from the text"""
    )
    named_entities: Optional[List[Any]] = Field(
        default_factory=list, description="""Named entities extracted from the text"""
    )


class NamedEntity(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Gene(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class GeneDescriptionTerm(NamedEntity):
    label: Optional[str] = Field(None, description="""the name of the GO term""")
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")


class CompoundExpression(ConfiguredBaseModel):
    None


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """

    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(
        None, description="""A qualifier for the statements, e.g. \"NOT\" for negation"""
    )
    subject_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""",
    )
    object_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""",
    )


class TextWithTriples(ConfiguredBaseModel):
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
GeneDescription.update_forward_refs()
ExtractionResult.update_forward_refs()
NamedEntity.update_forward_refs()
Gene.update_forward_refs()
GeneDescriptionTerm.update_forward_refs()
CompoundExpression.update_forward_refs()
Triple.update_forward_refs()
TextWithTriples.update_forward_refs()
RelationshipType.update_forward_refs()
Publication.update_forward_refs()
AnnotatorResult.update_forward_refs()
